#!/usr/bin/env python
# Copyright 2008-2010 David D Lowe
# Released under the GPLv3
"""Contains the Shell class and its subclasses"""

from __future__ import division
import const
from const import logger, MY_CACHE_HOME, MY_DATA_HOME, PIGMENT_TYPES
import pygtk
pygtk.require("2.0")
import gtk
import pexpect
import os, sys
import time
from ConfigParser import SafeConfigParser
import ConfigParser
from gettext import gettext as _
import subprocess
import dbus

DEBUG_SHELL = True
if not const.DEBUG:
    DEBUG_SHELL = False

def bash_prepare(string): 
    """Prepares a string for bash by surrounding it with single quotes.
    >>> shell.bash_prepare("two'words")
    "'two'\\''words'"
    """ 
    return "'" + string.replace("'", r"'\''") + "'"

class Shell():
    """An object which can run commands
    This is an "abstract" class, only its subclasses should be used"""
    def __init__(self):
        self.ready = False
    
    def prepare(self):
        """Prepares the shell, may ask user for password if necessary"""
        return self.ready
    
    def do(self, command, intrShell=False, cwd=None):
        """Runs the command in this shell.
        If intrShell is True, allows the shell to interpret special characters
        such as *
        cwd: the current directory for the command"""
        pass
    
    def set_out_file(self, outfile):
        """Sets the file where stdout output is put"""
        pass
    
    def exit(self):
        """Exits the shell"""
        self.ready = False

class RootShell(Shell):
    """A shell that runs commands as root. It will automatically use
    BashRootShell or PolicyKitShell in its backend depending on the
    user's settings"""
    
    def __init__(self):
        Shell.__init__(self)
        configParser = SafeConfigParser()
        configParser.read("/etc/epidermis.conf")
        try:
            val = configParser.get("epidermis","access_root")
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError),ee:
            print >> sys.stderr, "Error parsing /etc/epidermis.conf, using gksu"
            val = "gksu"
        if val == "policykit":
            self._shell = PolicyKitShell()
        elif val == "gksu":
            self._shell = BashRootShell()
        else:
            print >> sys.stderr, "Error parsing /etc/epidermis.conf, using gksu"
            self._shell = BashRootShell()
        
        def wrapper(func):
            def execute_and_update_ready(*args, **kwargs):
                func(*args, **kwargs)
                self.ready = self._shell.ready
            return execute_and_update_ready
        
        self.prepare = wrapper(self._shell.prepare)
        self.do = wrapper(self._shell.do)
        if hasattr(self._shell,"set_err_file"):
            self.set_err_file = wrapper(self._shell.set_err_file)
        self.set_out_file = wrapper(self._shell.set_out_file)
        self.exit = wrapper(self._shell.exit)
    
    

class BashRootShell(Shell):
    """The Shell object which has access to a bash terminal run as root
    It uses pexpect, and it runs sudo bash -i"""
    def __init__(self, prompt="epidermisbash:"):
        Shell.__init__(self)
        self.prompt = prompt
        self.outfile = os.tmpfile()
        self.gksubash, self.gksued = None, None
        self.spawn = None
        self.ready = False
        import handy
        self.gksubash = handy.get_data_file("gksubash.sh")
        self.gksued = handy.get_data_file("gksued.sh")
        if not os.path.exists(self.gksubash):
            raise(Exception("Cannot find gksubash.sh"))
        if not os.path.exists(self.gksued):
            raise(Exception("Cannot find gksued.sh"))
        
    def prepare(self):
        """Initializes the bash shell as root, will ask user for 
        password graphically if necessary
        Will raise a ShellNotReadyException if it fails to prepare itself
        # Unfortunately, due to bug #244930, gksu cannot run bash -i correctly
        # This workaround uses sudo, but still asks for the password
        # graphically, if required
        # See bug: https://bugs.launchpad.net/ubuntu/+source/gksu/+bug/244930"""
        self.spawn = pexpect.spawn("sh", [self.gksubash, self.gksued])
        self.spawn.logfile = self.outfile
        res = self.spawn.expect(["EPIDERMIS_PASS", "EPIDERMIS_SUDO_RUN_SUCCESSFULLY", \
            "EPIDERMIS_SUDO_FINISHED"])
        self.ready = True
        while res != 1: # sudo didn't run successfully
            if res == 2:
                # sudo exited unexpectedly
                self.ready = False
                raise(ShellNotReadyException())
                break
            # I need a password, use gksu gui to get one
            pop = os.popen("gksu -p --message " + bash_prepare(_("Enter password to allow Epidermis to access root privileges")))
            password = "IMPOSSIBLE_PASSWORD_ITHINK3457865"
            for line in pop.readlines():
                if "sn_launcher_context_complete" in line:
                    continue
                if len(line) == 0:
                    continue
                password = line[:-1]
            if password == "IMPOSSIBLE_PASSWORD_ITHINK3457865" or password == "":
                retry = self._yes_no_box(_("Could not retrieve password, try again?"))
                if not retry:
                    self.ready = False
                    raise(ShellNotReadyException())
                    break
                else:
                    continue
            # try this password, disable outfile to avoid printing password
            oldoutfile = self.spawn.logfile
            self.spawn.logfile = os.tmpfile()
            self.spawn.sendline(password)
            self.spawn.logfile = oldoutfile
            # recheck to see if successful
            res = self.spawn.expect(["EPIDERMIS_PASS", "EPIDERMIS_SUDO_RUN_SUCCESSFULLY", \
                "EPIDERMIS_SUDO_FINISHED"])
        
        self.spawn.readline()
        
        if not self.ready:
            raise(ShellNotReadyException())
        else:
            if hasattr(self, "outfile"):
                self.spawn.logfile = self.outfile
            ##self.spawn.sendline(" PS1=" + self.prompt)
            ##self.spawn.readline()
            ##self.spawn.expect(self.prompt)
            self.do(" PS1=" + self.prompt)
            self.do(" HISTCONTROL=\"ignoreboth\"")
            return True
    
    def set_out_file(self, outfile):
        """sets where the bash should print to, by default, no where
        outfile: file like object"""
        if not self.spawn is None:
            Sfile = outfile
        self.outfile = outfile
    
    
    def do(self, command, intrShell=False, cwd=None, expectStr=None, expectNum=1 ):
        """Runs the command
        command: the command string as bash would expect, or a list of command
        arguments
        expectStr: the string to stop execution at, if not specified, Shell will
        stop at bash prompt
        cwd: the current directory for the command run
        expectNum: if expectStr is specified, this number indicates the number
        of times do must expect expectStr
        intrShell: if False, do will automatically prepare bash on each of the
        command's list's items. Currently has no effect if command is str"""
        
        if not self.ready:
            raise(ShellNotReadyException())
            
        cmd = ""
        if isinstance(command, list):
            for item in command:
                if intrShell:
                    cmd = cmd + item.replace(" ", r"\ ") + " "
                else:
                    cmd = cmd + bash_prepare(item) + " "
            cmd = cmd[:-1]
        elif isinstance(command, str):
            cmd = command
        else:
            raise(Exception("command must be list or string"))
        
        if DEBUG_SHELL:
            logger.debug("debug+++root do: " + cmd)
        if not cwd is None:
            self.spawn.sendline(" cd " + bash_prepare(cwd))
            self.spawn.readline()
            self.spawn.expect(self.prompt)
        self.spawn.sendline(" " + cmd) # space is necessary so that the command is not 
                                       # recorded in history
        self.spawn.readline() # reads the line in which the command was entered
        if expectStr is None:
            self.spawn.expect(self.prompt)
        else:
            coun = 0
            while coun < expectNum:
                self.spawn.expect(expectStr)
                coun += 1
            
    
    def _yes_no_box(self, message):
        dd = gtk.MessageDialog(type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_YES_NO, \
            message_format=message)
        res = dd.run()
        dd.hide()
        dd.destroy()
        if res == gtk.RESPONSE_YES:
            return True
        else:
            return False
    
    def exit(self):
        """exits the bash shell, terminates it if necessary"""
        if self.spawn.isalive():
            self.do("exit", expectStr="exit")
        
        if self.spawn.isalive():
            time.sleep(0.3)
            if self.spawn.isalive():
                self.spawn.terminate()
        self.ready = False

class PolicyKitShell(Shell):
    """An object which can run commands using a DBus service and 
    PolicyKit authentication"""
    def __init__(self):
        self.ready = False
        self.outFile, self.errFile = None, None
        self.pshell = None
        
    def prepare(self):
        bus = dbus.SystemBus()
        self.pshell = bus.get_object("org.tuxfamily.epidermis.Shell",
            "/org/tuxfamily/epidermis/PShell")
        rt = self.pshell.prepare(dbus_interface="org.tuxfamily.epidermis.PShellInterface")
        self.ready = rt
        if not rt:
            raise(ShellNotReadyException())
        return self.ready
    
    def set_out_file(self, outfile):
        self.outFile = outfile
    
    def set_err_file(self, errfile):
        self.errFile = errfile
    
    def do(self, command, intrShell=False, cwd=None):
        """Runs the command.
        command: can only be a list
        If intrShell is True, allows the shell to interpret special characters
        such as *
        cwd: current directory"""
        if not isinstance(command, list):
            raise(Exception("Not list"))
        global DEBUG_SHELL
        if DEBUG_SHELL:
            cmdStr = ""
            for it in command:
                cmdStr = cmdStr + it + " "
            if len(cmdStr) > 0:
                cmdStr = cmdStr[:-1]
            logger.debug("debug+++pshell do: " + cmdStr)
        if cwd is None:
            cwd = "/tmp/"
        exitcode, out, err = self.pshell.do(command,intrShell,
            cwd, dbus_interface="org.tuxfamily.epidermis.PShellInterface")
        if not self.outFile is None:
            self.outFile.write(out)
        if not self.errFile is None:
            self.errFile.write(err)
        return exitcode
    
    def exit(self):
        self.pshell.leave(dbus_interface="org.tuxfamily.epidermis.PShellInterface")
        self.pshell = None
        self.ready = False


class SubprocessShell(Shell):
    """An object which can run commands using subprocess.call as current user"""
    def __init__(self):
        self.ready = True
        self.outFile = sys.stdout
        self.errFile = sys.stderr
    
    def prepare(self):
        """Prepares the shell"""
        self.ready = True
        self.outFile = sys.stdout
        return self.ready
    
    def do(self, command, intrShell=False, cwd=None):
        """Runs the command.
        command: can only be a list
        If intrShell is True, allows the shell to interpret special characters
        such as *
        cwd: current directory"""
        if not isinstance(command, list):
            raise(Exception("Not list"))
        global DEBUG_SHELL
        if DEBUG_SHELL:
            cmdStr = ""
            for it in command:
                cmdStr = cmdStr + it + " "
            if len(cmdStr) > 0:
                cmdStr = cmdStr[:-1]
            logger.debug("debug+++subprocess do: " + cmdStr)
        subprocess.call(command, shell=intrShell, cwd=cwd, stdout=self.outFile, \
            stderr=self.errFile)
        
    def set_out_file(self, outfile):
        self.outFile = outfile
    
    def set_err_file(self, errfile):
        self.errFile = errfile
    
    def exit(self):
        """Exits the shell, this shell is always ready so this method has no
        function"""
        self.ready = True

class ShellNotReadyException(Exception):
    def __str__(self):
        return "Shell has not been prepared"

def testshell():
    """A function that tries some of BashRootShell's and PolicyKitShell's
    functionality"""
    bashShell = BashRootShell()
    bashShell.set_out_file(sys.stdout)
    bashShell.prepare()
    if bashShell.ready == False:
        print >> sys.stderr, "bash shell is not ready, user cancelled?"
        return
    print "Bash shell is ready"
    bashShell.do(["whoami"])
    bashShell.exit()
    
    pkShell = PolicyKitShell()
    pkShell.set_out_file(sys.stdout)
    pkShell.prepare()
    if pkShell.ready == False:
        print >> sys.stderr, "policykit shell is not ready, user cancelled?"
        return
    print "PolicyKit shell is ready"
    pkShell.do(["whoami"])
    pkShell.exit()
    
    rtShell = RootShell()
    rtShell.set_out_file(sys.stdout)
    rtShell.prepare()
    if rtShell.ready == False:
        print >> sys.stderr, "root shell is not ready"
        return
    print "Root shell is ready"
    rtShell.do(["whoami"])
    rtShell.exit()

def determine_path ():
    """Borrowed from wxglade.py"""
    try:
        root = __file__
        if os.path.islink (root):
            root = os.path.realpath (root)
        return os.path.dirname (os.path.abspath (root))
    except Exception, ee:
        print >> sys.stderr, "I'm sorry, but something is wrong."
        print >> sys.stderr, "There is no __file__ variable. Please contact the author."
        sys.exit (1)
        raise(ee)


if __name__ == "__main__":
    print >> sys.stderr, "do not run"
    sys.exit(1)
